from app.utils import op_dev_service


def run(parser, args):
  if args.operation != 'get':
    return

  credentials = op_dev_service.call('get-code-commit-credentials-for-team', { 'team_id': args.team_id })
  print('username=%s' % credentials['user_name'])
  print('password=%s' % credentials['password'])
